"""
Helpers and utilities for the imdb_rating_classifier app.
"""
from __future__ import annotations

import json
import typing as t

import requests


def unpack_contents(response: requests.Response, cutoff: str = None) -> list[dict[str, t.Any]]:
    """
    Helper method to unpack the contents from request response,
    reporting errors in a helpful manner, if any.

    Args:
        response (requests.Response): request response.
        cutoff (str, optional): cutoff. Defaults to None.

    Returns:
        list[dict[str, t.Any]]: list of dictionaries.
    """
    try:
        if response.status_code == 200 and cutoff is None:
            return response
        elif response.status_code == 200 and cutoff == 'dict':
            return response.json()
        else:
            raise Exception(
                f'Exception: {response.status_code} - {response.reason}',
                f'Traceback: {response.text}',
            )

    except json.decoder.JSONDecodeError as err:
        raise Exception(f'Error: {err}') from err



"""
Application entry point for generating the dataset.
"""
from __future__ import annotations

import os
import sys
from concurrent.futures import ThreadPoolExecutor

import click
import pandas as pd

sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
from imdb_rating_classifier import __version__  # noqa: E402
from imdb_rating_classifier.penalizer import penalize_reviews  # noqa: E402
from imdb_rating_classifier.schema import MovieChart, normalize  # noqa: E402
from imdb_rating_classifier.scraper import Scraper, get_movie_oscar_data, logger  # noqa: E402

# help context
CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])
# get the base directory
BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))


@click.group(
    invoke_without_command=True,
    context_settings=CONTEXT_SETTINGS,
    help='Application entry point for IMDB rating classifier.',
)
@click.version_option(version=__version__, prog_name='IMDB Rating Classifier')
@click.pass_context
def main(ctx: click.Context) -> None:
    """
    Application entry point for IMDB rating classifier.

    Args:
        ctx (click.Context): The click context.
    """
    if ctx.invoked_subcommand is None:
        print(ctx.get_help())


@click.command()
@click.option(
    '--output',
    type=click.Path(exists=False, file_okay=True, dir_okay=False),
    default='movies.csv',
    help='The path to the output file.',
)
@click.option(
    '--number-of-movies',
    type=click.INT,
    default=25,
    help='The number of movies to scrape.',
)
def generate(output: str, number_of_movies: int) -> None:
    """
    Generate the output dataset containing both
    the original and adjusted ratings.

    An extra JSON file will be generated alongside the csv file
    """
    # scrape IMDB movie chart data
    scraper = Scraper(
        url='https://www.imdb.com/chart/top',
        number_of_movies=number_of_movies,
    )
    movies = scraper.scrape()

    # get the oscars winners data
    logger.info('Getting the Oscar winners data...')
    # speed up the process
    with ThreadPoolExecutor(max_workers=10) as executor:
        results = executor.map(get_movie_oscar_data, [movie['url'] for movie in movies])
    # add the oscars data to the movies
    for movie, result in zip(movies, results):
        movie['oscars_won'] = result

    # convert the original set of movies to dataframe
    logger.info('Converting to original set of movies to a dataframe...')
    raw_input_df = pd.DataFrame(movies)

    # normalize the dataframe
    logger.info('Normalizing the dataframe...')
    refined_df = normalize(raw_input_df)

    # recreate the movies dict from the dataframe
    movies_dict = refined_df.to_dict(orient='records')

    # validate the movies
    logger.info('Validating the movies...')
    try:
        valid_movies = [MovieChart(**movie) for movie in movies_dict]
    except Exception as e:
        logger.error(e)
        raise
    logger.info(F'Data validation passed. The number of valid movies: {len(valid_movies)}')

    # penalize the movies based on the ruleset defined in the penalizer module
    logger.info('Penalizing movies...')
    penalized_movies = penalize_reviews(movies_dict)
    # check the number of penalized movies
    number_of_penalized_movies = len(
        [movie for movie in penalized_movies if 'penalized' in movie and movie['penalized'] is True],
    )
    logger.info(f'the number of panalized movies: {number_of_penalized_movies}')

    # convert the penalized set of movies to dataframe
    logger.info('Converting the set of penalized movies to a dataframe...')
    penalized_df = pd.DataFrame(penalized_movies)

    # sort the dataframe by rank in a descending order
    penalized_df.sort_values(by='rank', ascending=False, inplace=True)
    print(penalized_df.head())

    # save to csv
    logger.info('Saving the dataset to csv...')
    penalized_df.to_csv(output, index=False)

    # save to JSON
    logger.info('Saving a copy to JSON...')
    penalized_df.to_json(output.replace('.csv', '.json'), orient='records', indent=2)
    logger.info('Done!')


# add commands to main
main.add_command(generate)


if __name__ == '__main__':
    raise SystemExit(main())  # pragma: no cover



"""
Module for predefined table/JSON schemas

The intent of this module is to define the schema for the IMDB movie chart data.
@TODO: - Validate the schema using a dataclass?
"""
from __future__ import annotations

import typing as t
from dataclasses import dataclass

import pandas as pd


@dataclass
class MovieChart:
    """
    Movie chart data class

    Validate the schema using a dataclass?
    """

    rank: int
    title: str
    year: int
    rating: float
    votes: int
    url: str
    oscars_won : int
    penalized: bool

    def __post_init__(self):
        """
        Post initialization
        """
        # validate that the rank is between 1 and 250
        if not 1 <= self.rank <= 250:
            raise ValueError(f'Invalid rank: {self.rank}')

        # validate that the year is between 1900 and 2023
        if not 1900 <= self.year <= 2023:
            raise ValueError(f'Invalid year: {self.year}')

        # validate that the rating is between 0.0 and 10.0
        if not 0.0 <= self.rating <= 10.0:
            raise ValueError(f'Invalid rating: {self.rating}')

        # validate that the votes is greater than 0
        if not self.votes > 0:
            raise ValueError(f'Invalid votes: {self.votes}')

        # validate that the penalized is a boolean
        if not isinstance(self.penalized, bool):
            raise ValueError(f'Invalid penalized: {self.penalized}')

        # validate that the oscars won is an integer
        if not isinstance(self.oscars_won, int):
            raise ValueError(f'Invalid oscars_won: {self.oscars_won}')

    def __repr__(self) -> str:
        """
        Movie chart representation
        """
        return f'MovieChart({self.to_dict()})'

    def to_dict(self) -> dict[str, t.Any]:
        """
        Convert the MovieChart object to a dictionary

        Returns:
            dict: Movie chart data.
        """
        return {
            'rank': self.rank,
            'title': self.title,
            'year': self.year,
            'rating': self.rating,
            'votes': self.votes,
            'url': self.url,
            'oscars_won': self.oscars_won,
            'penalized': self.penalized,
        }


def normalize(movies_df: pd.DataFrame) -> pd.DataFrame:
    """
    Cast the Dataframe columns to the correct type.

    Args:
        movies_data (pd.DataFrame): Movie chart data.

    Returns:
        pd.DataFrame: Movie chart data.
    """
    # cast the year column to int
    movies_df['year'] = movies_df['year'].astype(int)

    # cast the rating column to float
    movies_df['rating'] = movies_df['rating'].astype(float)

    # cast the votes column to int
    movies_df['votes'] = movies_df['votes'].astype(int)

    # cast the rank column to int
    movies_df['rank'] = movies_df['rank'].astype(int)

    # cast the penalized column to bool
    movies_df['penalized'] = movies_df['penalized'].astype(bool)

    return movies_df



"""
Module for review penalizer processing logic

Review Penalizer:
Ratings are good because they give us an impression of how many people think a movie is good or bad.
However, it does matter how many people voted.
The aim is to penalize those movies where the number of ratings is low.

Find the movie with the maximum number of ratings (out of the TOP 20 only).
This is going to be the benchmark. Compare every movie’s number of ratings to this
and penalize each of them based on the following rule:
- Every 100k deviation from the maximum translates to a point deduction of 0.1.

For example:
-   suppose that the movie with the highest number of ratings has 2.456.123 ratings.
    This means that for a given movie with 1.258.369 ratings and an IMDB review score of 9.4,
    the amount of the deduction is 1.1 and therefore the adjusted rating is 8.3.

@TODO: [X] - fix the type hinting for the movies parameter
       [X] - fix the logic for the numbber of ratings lookup
       (currently we are getting the higher review instead of the higher number of ratings)
"""
from __future__ import annotations

from typing import Any


def penalize_reviews(movies: list[dict[str, Any]]) -> list[dict[str, Any]]:
    """
    penalize reviews form a list of movies

    Args:
        movies (list[dict[str, Any]]): list of movies.

    Returns:
        list[dict[str, Any]]: list of penalized movies.
    """
    # get max number of reviews
    max_reviews = max(movie['votes'] for movie in movies)
    # get penalized reviews
    for movie in movies:
        # if the number of votes is None, raise an exception
        if movie['votes'] is None:
            raise ValueError('The number of reviews cannot be None or 0')
        # get the number of reviews
        else:
            review = movie['votes']
            # if the number of reviews is less than the max number of reviews then penalize
            # else keep the rating as is and set penalized to False
            if review == max_reviews:
                movie['penalized'] = False
                movie['penalized_rating'] = None
            elif review < max_reviews:  # be explicit
                # get number of reviews difference
                reviews_difference = max_reviews - review
                # get the points deduction
                points_deduction = reviews_difference / 100000 * 0.1
                # get the penalized rating and round it to 1 decimals
                penalized_rating = round(movie['rating'] - points_deduction, 1)
                # update penalized review
                movie['penalized_rating'] = penalized_rating
                movie['penalized'] = True

        # re-penalize reviews based on the oscars awards
        movie = oscar_calculation(movie)

    return movies


def oscar_calculation(movies: dict[str, Any]) -> dict[str, Any]:
    """
    re-penalize reviews based on the oscars awards

    (pseudo code)
    >>> if penalized_rating is None and oscars_won >= 0: penalized_rating = None
    >>> for 1 or 2 oscars: penalized_rating = penalized_rating + 0.3
    >>> for 3 to 5 oscars: penalized_rating = penalized_rating + 0.5
    >>> for 6 to 10 oscars: penalized_rating = penalized_rating + 1.0
    >>> for 11 or more oscars: penalized_rating = penalized_rating + 1.5
    """
    if movies['penalized'] is False:
        pass

    else:
        if movies['oscars_won'] == 0:
            return movies

        elif 1 <= movies['oscars_won'] <= 2:
            movies['penalized_rating'] += 0.3
            return movies

        elif 3 >= movies['oscars_won'] <= 5:
            movies['penalized_rating'] += 0.5
            return movies

        elif 6 >= movies['oscars_won'] <= 10:
            movies['penalized_rating'] += 1.0
            return movies

        elif movies['oscars_won'] >= 11:
            movies['penalized_rating'] += 1.5
            return movies



"""
Module for scraping IMDB movie chart data.

The intent of this module is to scrape the IMDB movie chart data (TOP 250)
https://www.imdb.com/chart/top/ and return a list of movie data objects.

The movie data objects are defined in the schema module. We will design the data structure to be
compatible with the schema module.

The IMDB movie chart data is scraped using the requests library. The data is then parsed using the
BeautifulSoup library.

@TODO: [X] - Add support for scraping the number of reviews given to a movie.
       [X] - Add support for scraping the oscars won by a movie.
"""
from __future__ import annotations

import logging

import requests
from bs4 import BeautifulSoup

from imdb_rating_classifier.util.unpack import unpack_contents

# initialize logger
logger = logging.getLogger('imdb_rating_classifier')
logger.setLevel(logging.INFO)
formatter = logging.Formatter(
    '%(asctime)s:%(name)s:%(levelname)s:%(message)s',
    datefmt='[%Y-%m-%d][%H:%M:%S]',
)
# prints the log message to the console
stream_handler = logging.StreamHandler()
stream_handler.setFormatter(formatter)
logger.addHandler(stream_handler)

# html tags and attributes
IMDB_URL = 'https://www.imdb.com/chart/top/'
# the IMDB movie chart data is stored in a table with the class "lister-list"
IMDB_MOVIE_CHART_SELECTOR = 'tbody.lister-list tr'
IMDB_MOVIE_RANK_SELECTOR = 'td.titleColumn'
IMDB_MOVIE_TITLE_SELECTOR = 'td.titleColumn a'
IMDB_MOVIE_YEAR_SELECTOR = 'td.titleColumn span.secondaryInfo'
IMDB_MOVIE_RATING_SELECTOR = 'td.ratingColumn strong'  # "title" attribute of the "a" tag


class Scraper:
    """
    Class for scraping IMDB movie chart data.

    @TODO: - Add support for scraping multiple pages of the IMDB movie chart?
           - static typing to the class?
    """

    def __init__(self, number_of_movies: int, url: str = IMDB_URL):
        """
        Initialize Scraper class.

        Args:
            url (str, optional): IMDB movie chart URL. Defaults to IMDB_URL.
        """
        self.url = url
        self.number_of_movies = number_of_movies

    def scrape(self) -> list[dict]:
        """
        Scrape IMDB movie chart data and return a list of movie data objects.

        Returns:
            list[dict]: List of movie data objects.
        """
        # initialize movie chart data
        movie_chart_data = []

        # scrape IMDB movie
        logger.info('Scraping IMDB movie chart data...')
        response = unpack_contents(response=requests.get(self.url))
        # TODO: redundant code..remove?
        response.raise_for_status()

        # parse IMDB movie chart data
        soup = BeautifulSoup(response.text, 'html.parser')
        movie_chart = soup.select(IMDB_MOVIE_CHART_SELECTOR)

        # parse movie chart data
        logger.info('Parsing IMDB movie chart data...')
        for movie in movie_chart:
            # parse the rank of the movie and remove the new line characters.
            movie_rank = movie.select_one(IMDB_MOVIE_RANK_SELECTOR).text.strip().split('.')[0]
            movie_title = movie.select_one(IMDB_MOVIE_TITLE_SELECTOR).text.strip()
            movie_year = movie.select_one(IMDB_MOVIE_YEAR_SELECTOR).text.strip()
            movie_rating = movie.select_one(IMDB_MOVIE_RATING_SELECTOR).text.strip()
            movie_votes = movie.select_one(IMDB_MOVIE_RATING_SELECTOR).get('title').split()[3]
            movie_url = movie.select_one(IMDB_MOVIE_TITLE_SELECTOR).get('href')

            # create a movie data object
            movie_data = {
                'rank': movie_rank,
                'title': movie_title,
                'year': movie_year,
                'rating': movie_rating,
                'votes': movie_votes,
                'url': movie_url,
                'penalized' : False,
            }
            # append movie data object to movie chart data
            movie_chart_data.append(movie_data)

        # clean movie chart data
        movie_chart_data = self.clean_movie_chart_data(movie_chart_data)

        return movie_chart_data[: self.number_of_movies]

    @staticmethod
    def clean_movie_chart_data(movie_chart_data: list[dict]) -> list[dict]:
        """
        Clean movie chart data.

        Args:
            movie_chart_data (list[dict]): List of movie data objects.

        Returns:
            list[dict]: List of cleaned movie data objects.
        """
        # clean movie chart data
        logger.info('Cleaning IMDB movie chart data...')
        for movie in movie_chart_data:
            # remove the parentheses from the movie year
            movie['year'] = movie['year'].replace('(', '').replace(')', '')
            # remove the comma from the number of votes
            movie['votes'] = movie['votes'].replace(',', '')
            # remove the leading slash from the movie url and append the IMDB url
            movie['url'] = movie['url'].replace('/', '', 1)
            movie['url'] = f'https://www.imdb.com/{movie["url"]}'

        return movie_chart_data


def get_movie_oscar_data(movie_url: str) -> int:
    """
    Get the Oscar data for a movie.

    Args:
        movie_url (str): IMDB movie URL.

    Returns:
        int: Number of Oscars won by the movie.
    """
    result = requests.get(
        movie_url,
        headers={
            'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_13_6) \
                        AppleWebKit/537.36 (KHTML, like Gecko) \
                        Chrome/70.0.3538.77 Safari/537.36',
        },
    )
    content = result.text
    # parse the html content
    more_soup = BeautifulSoup(content, 'html.parser')
    element = more_soup.find_all(class_='ipc-metadata-list-item__label')

    # parse movie Oscar data
    try:
        extracted_text = element[6].get_text(strip=True).split()
        if extracted_text[0] == 'Won' and 'Oscar' in extracted_text[-1]:
            return int(extracted_text[1])

    except IndexError:
        pass

    return 0
